@@ -14,6 +14,7 @@
 
 
 import sys
+import os
 if "darwin" in sys.platform.lower():
     # To avoid multiple runs of the model code
     # https://pythonspeed.com/articles/python-multiprocessing/
@@ -47,9 +48,51 @@ def version_check(version):
         return False
 
 
-def init_process_group(backend):
+def init_process_group(backend,
+                       init_method=None,
+                       world_size=None,
+                       rank=None):
+    """
+    Initializes the default distributed process group and the AdaptDL
+    collectives module.
+
+    Args:
+        backend (str or Backend): The backend to use. Use "nccl" for multi-GPU
+            training else "gloo".
+        init_method (str, optional): URL specifying how to initialize the
+                                     process group.
+        world_size (int, optional): Number of processes participating in
+                                    the job
+        rank (int, optional): Rank of the current process (it should be a
+                              number between 0 and ``world_size``-1).
+
+    If init_method, world_size and rank is NOT provided, typically in the
+    Kubernetes environment, AdaptDL will try to infer them through environment
+    variables ADAPTDL_MASTER_ADDR, ADAPTDL_NUM_REPLICAS and
+    ADAPTDL_REPLICA_RANK respectively.
+    """
+    if adaptdl.env.from_ray():
+        from adaptdl_ray.adaptdl.utils import unique_nodes_pg
+        assert init_method is not None
+        assert world_size is not None
+        assert rank is not None
+        os.environ["ADAPTDL_NUM_NODES"] = str(unique_nodes_pg())
+        os.environ["ADAPTDL_REPLICA_RANK"] = str(rank)
+        os.environ["ADAPTDL_NUM_REPLICAS"] = str(world_size)
+
     url = adaptdl.env.supervisor_url()
-    if url:
+    master_port = adaptdl.env.master_port()
+    if rank is None:
+        rank = adaptdl.env.replica_rank()
+
+    if world_size is None:
+        world_size = adaptdl.env.num_replicas()
+
+    if init_method is not None:
+        _, master_addr, master_port = init_method.split(":")
+        master_addr = master_addr[2:]
+        master_port = int(master_port)
+    elif url:
         key = adaptdl.env.job_id()
         group = adaptdl.env.num_restarts()
         while True:
@@ -70,16 +113,16 @@ def init_process_group(backend):
     else:
         master_addr = adaptdl.env.master_addr()
 
-    master_port = adaptdl.env.master_port()
-
     # Initialize collective module.
-    adaptdl.collective.initialize(master_addr, master_port)
+    adaptdl.collective.initialize(master_addr,
+                                  master_port,
+                                  rank,
+                                  world_size)
 
     # Initialize torch.distributed.
     torch_port = adaptdl.collective.broadcast(portpicker.pick_unused_port())
     init_method = "tcp://{}:{}?rank={}&world_size={}".format(
-            master_addr, torch_port, adaptdl.env.replica_rank(),
-            adaptdl.env.num_replicas())
+            master_addr, torch_port, rank, world_size)
     LOG.info("Initializing torch.distributed using %s", init_method)
     torch.distributed.init_process_group(backend, init_method)
 
