@@ -14,18 +14,26 @@
 
 
 import functools
+import numpy as np
 import time
+import warnings
+from typing import Optional
 
 import torch
 import torch.cuda
+import torch.distributed
 from torch.autograd import Variable
 from torch.nn.parallel import DistributedDataParallel
 
 import adaptdl.checkpoint
 import adaptdl.env
+import adaptdl.utils
 from adaptdl.torch.data import current_dataloader
-from adaptdl.torch.adascale import AdaScale
-from adaptdl.torch._metrics import profile_sync_time
+from adaptdl.torch.scaling_rules import AdaScale, AdamScale, ScalingRuleBase
+from adaptdl.torch.gradient_noise_scale import GradientNoiseScale,\
+                                               AdamGradientNoiseScale
+from adaptdl.torch._metrics import profile_sync_time, update_grad_params,\
+                                   update_progress
 
 
 class AdaptiveDataParallel(DistributedDataParallel):
@@ -34,19 +42,22 @@ class AdaptiveDataParallel(DistributedDataParallel):
     adaptive batch sizes and checkpoint-restart elasticity. It automatically
     saves the given model, optimizer, and (optionally) LR scheduler whenever a
     checkpoint is triggered, and restores their states after restart. The
-    optimizer is automatically patched with AdaScale.
+    optimizer is automatically patched with the chosen scaling rule.
 
     Arguments:
         model (torch.nn.Module): Model to be distributed.
         optimizer (torch.optim.Optimizer): Optimizer used to update the given
-            model's parameters, will be patched using
-            :class:`adaptdl.torch.adascale.AdaScale`.
+        model's parameters, will be patched using subclass of
+        :class:`adaptdl.torch.scaling_rules.ScalingRuleBase`.
+        scaling_rule (ScalingRuleBase): Scaling rule used to
+        patch the given optimizer, default to AdaScale.
         lr_scheduler (torch.optim.lr_scheduler._LRScheduler): LR scheduler used
-            to anneal the learning rate for the given optimizer.
+        to anneal the learning rate for the given optimizer.
         name (string): Unique name for each instance of this class, needed only
-            if multiple instances exist.
+        if multiple instances exist.
     """
-    def __init__(self, model, optimizer, lr_scheduler=None,
+    def __init__(self, model, optimizer, lr_scheduler=None, mp_scaler=None,
+                 scaling_rule: Optional[ScalingRuleBase] = None,
                  name="adaptdl-dataparallel", **kwargs):
         super().__init__(model, **kwargs)
         self._key = id(self)
@@ -58,30 +69,38 @@ class AdaptiveDataParallel(DistributedDataParallel):
         for param in model.parameters():
             param.register_hook(functools.partial(self._backward_hook, param))
 
-        # Setup for AdaScale, must be after registering backward hooks!
-        self.adascale = AdaScale(optimizer, patch_optimizer=True)
+        # Setup for the scaling_rule, must be after registering backward hooks
+        # because some of them need to register their own backward hooks.
+        if not scaling_rule and (isinstance(optimizer, torch.optim.Adam) or
+                                 isinstance(optimizer, torch.optim.AdamW)):
+            self.scaling_rule = AdamScale()
+        else:
+            self.scaling_rule = scaling_rule or AdaScale()
 
-        self._state = _AdaptiveDataParallelState(model, optimizer,
-                                                 lr_scheduler, name)
+        if isinstance(scaling_rule, AdamScale):
+            self.gns = AdamGradientNoiseScale(self, optimizer,
+                                              mp_scaler=mp_scaler)
+        else:
+            self.gns = GradientNoiseScale(self, optimizer, mp_scaler=mp_scaler)
+        self.scaling_rule.initialize(self, optimizer, patch_optimizer=True)
+
+        self._state = _AdaptiveDataParallelState(
+            model, optimizer, lr_scheduler, mp_scaler, name)
         adaptdl.checkpoint.load_state(self._state)
 
         self._sync_start = None
 
     def forward(self, *args, **kwargs):
-        # Do not do gradient synchronization during gradient accumulation
-        # Otherwise, exactly the same as DistributedDataParallel's forward
+        # Do not do gradient synchronization during gradient accumulation.
         dataloader = current_dataloader()
-        accumulation_steps = dataloader.accumulation_steps
-        # TODO: move this to the dataloader.__iter__
-        self.adascale.set_accumulation_steps(accumulation_steps)
-        if (self.adascale.is_accumulation_step()):
-            with super().no_sync():
-                dataloader.is_accumulation_step = True
-                return super().forward(*args, **kwargs)
-        else:
-            dataloader.is_accumulation_step = False
-            return super().forward(*args, **kwargs)
+        if dataloader is not None and dataloader.training:
+            self.require_backward_grad_sync = dataloader.is_optim_step()
+            accum_scale = (dataloader.current_local_bsz *
+                           adaptdl.env.num_replicas() / dataloader.batch_size)
+            self.gns.set_accum_scale(accum_scale)
+        return super().forward(*args, **kwargs)
 
+    @adaptdl.utils.print_exc
     def _backward_hook(self, param, grad):
         # This method should be invoked once for each parameter during the
         # backward pass, before gradients are synchronized between replicas.
@@ -93,6 +112,7 @@ class AdaptiveDataParallel(DistributedDataParallel):
         self._final_callback_queued = False
         Variable._execution_engine.queue_callback(self._queue_callback)
 
+    @adaptdl.utils.print_exc
     def _queue_callback(self):
         # This method should be invoked after the entire backward pass. We want
         # to make sure self._final_callback is invoked once, only after all
@@ -106,6 +126,7 @@ class AdaptiveDataParallel(DistributedDataParallel):
         self._final_callback_queued = True
         Variable._execution_engine.queue_callback(self._final_callback)
 
+    @adaptdl.utils.print_exc
     def _final_callback(self):
         # This method should be invoked once for each backward pass, after
         # gradients have been synchronized between each replica.
@@ -132,20 +153,18 @@ class AdaptiveDataParallel(DistributedDataParallel):
         dataloader.train()
 
         scale = dataloader.current_batch_size / dataloader.batch_size
-        self.adascale.set_scale(scale)
-        self._state.gain = self.adascale.gain()
-        adaptdl.torch._metrics.update_progress(self.adascale.get_progress())
+        self._state.gain = self.gns.gain(scale)
+        self._state.lr_factor = \
+            np.average(self.scaling_rule.scale_lr(scale))
+        update_progress(self.gns.get_progress())
         if dataloader.max_batch_size and \
                 dataloader.max_batch_size > dataloader.batch_size:
-            adaptdl.torch._metrics.update_grad_params(
-                self._key, self.adascale.norm_avg(), self.adascale.var_avg())
+            update_grad_params(self._key, self.gns.sqr_avg(),
+                               self.gns.var_avg())
         self._sync_start = None
 
     def zero_grad(self, *args, **kwargs):
-        if self.adascale.is_accumulation_step():
-            return
-        else:
-            return super().zero_grad(*args, **kwargs)
+        warnings.warn("zero_grad has no effect with AdaptiveDataParallel")
 
     @property
     def gain(self):  # TODO: should be tracked in the metrics module instead.
@@ -167,32 +186,47 @@ class AdaptiveDataParallel(DistributedDataParallel):
         if tag_prefix and not tag_prefix.endswith("/"):
             tag_prefix += "/"
         writer.add_scalar(tag_prefix + "Gradient_Norm_Sqr",
-                          self.adascale.norm_avg(), global_step)
+                          self.gns.sqr_avg(), global_step)
         writer.add_scalar(tag_prefix + "Gradient_Variance",
-                          self.adascale.var_avg(), global_step)
+                          self.gns.var_avg(), global_step)
+        writer.add_scalar(tag_prefix + "Gain",
+                          self._state.gain, global_step)
         writer.add_scalar(tag_prefix + "Learning_Rate_Factor",
-                          self.adascale.gain(), global_step)
+                          self._state.lr_factor, global_step)
 
 
 class _AdaptiveDataParallelState(adaptdl.checkpoint.State):
-    def __init__(self, model, optimizer, lr_scheduler,
+    def __init__(self, model, optimizer, lr_scheduler, mp_scaler,
                  name="adaptdl-dataparallel"):
         super().__init__(name)
         self.model = model
         self.optimizer = optimizer
         self.lr_scheduler = lr_scheduler
+        self.mp_scaler = mp_scaler
         # TODO: Gain/goodput should be tracked in the metrics module instead.
-        self.gain = 0.0
+        self.gain = 1.0
+        # lr_factor summary
+        self.lr_factor = 1.0
 
     def save(self, fileobj):
         state_dicts = [self.model.state_dict(), self.optimizer.state_dict()]
+
         if self.lr_scheduler is not None:
             state_dicts.append(self.lr_scheduler.state_dict())
-        torch.save((state_dicts, self.gain), fileobj)
+        else:
+            state_dicts.append(None)
+
+        if self.mp_scaler is not None:
+            state_dicts.append(self.mp_scaler.state_dict())
+        else:
+            state_dicts.append(None)
+        torch.save((state_dicts, self.gain, self.lr_factor), fileobj)
 
     def load(self, fileobj):
-        state_dicts, self.gain = torch.load(fileobj)
+        state_dicts, self.gain, self.lr_factor = torch.load(fileobj)
         self.model.load_state_dict(state_dicts[0])
         self.optimizer.load_state_dict(state_dicts[1])
-        if len(state_dicts) > 2:
+        if state_dicts[2] is not None:
             self.lr_scheduler.load_state_dict(state_dicts[2])
+        if state_dicts[3] is not None:
+            self.mp_scaler.load_state_dict(state_dicts[3])
